package net.lenkaspace.robotForaging;

import java.awt.Dimension;
import java.awt.Point;

import net.lenkaspace.creeper.CRController;
import net.lenkaspace.creeper.model.CRSettings;
import net.lenkaspace.creeper.report.CRTimeSeriesReport;
import net.lenkaspace.robotForaging.images.ImageProvider;
import net.lenkaspace.robotForaging.model.Parameters;
import net.lenkaspace.robotForaging.model.Settings;
import net.lenkaspace.robotForaging.model.World;
import net.lenkaspace.robotForaging.view.ControlPanel;

import org.jfree.data.Range;

public class RobotForaging {
	
	private CRController crController;
	
	public RobotForaging(boolean startSimulation_, double speed_) {
		
		Parameters paremeters = Parameters.getSingleton();
		CRSettings.getSingleton().setShouldPrintGraphicReports(false);
		
		//---- create CRController
		crController = new CRController("Understanding the Role of Recruitment in Robot Foraging");
		
		//---- create image provider
		ImageProvider imageProvider = new ImageProvider();
		crController.setImageProvider(imageProvider);
		
		//---- optionally setup new world and renderer size...
		World world = new World(crController);
		crController.setWorld(world);
		
		//---- setup reports
		CRTimeSeriesReport report1 = new CRTimeSeriesReport(Settings.REPORT_RESOURCE_COLLECTED, new String[] {Settings.VAR_TOTAL_RESOURCE, Settings.VAR_RECRUITERD_RESOURCE});
		report1.setYAxisDisplayRange(new Range(0,1));
		crController.getReportController().addReport(Settings.REPORT_RESOURCE_COLLECTED, report1);
		
		for (int i=0; i<paremeters.numOfRobots; i++ ) {
			CRTimeSeriesReport rep1 = new CRTimeSeriesReport(Settings.REPORT_ROBOT + i, new String[] {Settings.VAR_DEPO_QUALITY, Settings.VAR_DEPO_EE, Settings.VAR_DEPO_ESTIM_EE, Settings.VAR_SEARCH_TIME, Settings.VAR_SEARCH_TIME_PERCENT, Settings.VAR_DISTANCE_FROM_BASE, Settings.VAR_DISTANCE_TRAVELLED, Settings.VAR_ODOM_ERROR});
			rep1.setIsScatterPlot(true);
			rep1.setShouldOutputVariableNames(true);
			crController.getReportController().addReport(Settings.REPORT_ROBOT + i, rep1);
		}
	
		//-- ... or optionally create new control panel that extends from CRControlPanel and set it
		if (!CRSettings.getSingleton().getIsConsoleOnlyBuild()) {
			ControlPanel controlPanel = new ControlPanel(crController.getControlPanel().getSize(), crController);
			crController.setControlPanel(controlPanel);
			
			String helpText = "";
			helpText += "********************************************************************************************************************\n";
			helpText += "************   UNDERSTANDING THE ROLE OF RECRUITMENT IN ROBOT FORAGING  ***********\n";
			helpText += "********************************************************************************************************************\n";
			helpText += "\n";
			helpText += "Simulation of foraging using a swarm of individualists (I-Swarm) and robots that recruit each other\n";
			helpText += "to resource deposits (B-Swarm).\n";
			helpText += "\n";
			helpText += "For more info visit http://lenkaspace.net/lab/swarmSystems/recruitmentInRobotForaging";
			helpText += "\n\n";
			helpText += "\n";
			helpText += "START SIMULATION\n";	
			helpText += "Choose the world, the number of deposits, number and variance of quality groups,\n";
			helpText += "number of robots and whether to use B-Swarm. Press START SIMULATION.\n";
			helpText += "\n";
			helpText += "SCENARIO TYPES:\n";
			helpText += "Random: Deposits scattered based on uniform random distribution\n";
			helpText += "Random distant: Deposits scattered based on uniform random distribution, at least\n";
			helpText += "                         1500 units away from the base\n";
			helpText += "Patchy: Deposits grouped by 10 into randomly placed patches, at least 500 units\n";
			helpText += "            away from the base.\n";
			helpText += "\n";
			helpText += "DEPOSIT QUALITY GROUPS\n";
			helpText += "When 0 quality groups, each deposit has a quality Q of 0.5. As groups are added, the\n";
			helpText += "centre of each group has Q=1 and the quality of nearby deposits decreases with their\n";
			helpText += "distance from the group centre. Use quality group variance to specify how far deposits\n";
			helpText += "are affected. Standard values range from 0.01 - 0.1.\n";
			helpText += "\n";
			helpText += "ROBOT COLOURS\n";
			helpText += "GREY = normal, RED = avoiding, BLUE = doing neighbourhood search, BLACK =  loading / unloading resource";
			helpText += "GREY WITH BLACK DOT = going to base to unload, LIGHT BLUE = going towards deposit, GREEN = near base walk\n";
			helpText += "\n";
			helpText += "NAVIGATION\n";
			helpText += "The world is 4000x4000 large and the camera needs to be moved in order to see hidden\n";
			helpText += "parts. Use the movement button on the control panel or the arrow keys to move.\n";
			CRSettings settings = CRSettings.getSingleton();
			Dimension size = new Dimension(600, 600);
			controlPanel.getHelpPopup().setSize(size);
			controlPanel.getHelpPopup().setText(helpText);
			controlPanel.getHelpPopup().setLocation(new Point((int)(settings.getWindowSize().width/2 - size.width/2),(int)(settings.getWindowSize().height/2- size.height/2 - 40)));
		}		
		
		if (startSimulation_) {
			crController.startSimulation(speed_, true); 
		}
	}
	
	/**
	 * Main function
	 */
	public static void main(String[] args) {
		//-- set parameters either based on command line or manually
		Parameters parameters = Parameters.getSingleton();
		CRSettings settings = CRSettings.getSingleton();
		
		parameters.reportFolderName = "myReport";
		parameters.numOfRuns = 1;
		parameters.trialDuration = 600; // 10 mins
		parameters.scenario = World.RANDOM;
		parameters.totalResourceAvailable = 1;
		parameters.odomSegmentLength = 1;
		parameters.numOfRobots = 20;
		parameters.useBSwarm = true;
		parameters.useNeighborhoodSearch = true;
		
		//-- environment params
		parameters.numOfDeposits = 100;
		parameters.numOfDepositQualityGroups = 0;
		parameters.depositQualityVariance = 0;
		
				
		//-- robot examined params
		parameters.odomVarianceCoeff = 0.005;
		parameters.communicationRadius = 60;
		parameters.timeToRandWalk = 120; //seconds
		parameters.timeToWalkNearBase = 120;
		
		//-- optimised params
		parameters.neighbourhoodSearchRadius = 120;
		parameters.neighbourhoodSearchTime = 10;
		parameters.nearDistanceToBase = 200;
		
		double speed = 0.75;
		boolean startSimulation = false;
				
				
		//-- set params from command line
		boolean consoleOnly = false;
		int counter = 0;
		for (String s: args) {
			switch (counter) {
			case 0:
				consoleOnly = Boolean.valueOf(s);
			case 1:
				parameters.reportFolderName = s;
				break;
			case 2:
				parameters.numOfRuns =  Integer.valueOf(s);
				break;
			case 3:
				parameters.numOfRobots = Integer.valueOf(s);
				break;
			//-- environment params
			case 4:
				System.out.println(s);
				if (s.equalsIgnoreCase("random")) {
					parameters.scenario = World.RANDOM;
				} else if (s.equalsIgnoreCase("randomDistant")) {
					parameters.scenario = World.RANDOM_DISTANT;
				} else if (s.equalsIgnoreCase("patchy")) {
					parameters.scenario = World.PATCHY;
				}
				break;
			case 5:
				parameters.numOfDeposits = Integer.valueOf(s);
				break;
			case 6:
				parameters.numOfDepositQualityGroups = Integer.valueOf(s);
				break;
			case 7:
				parameters.depositQualityVariance = Double.valueOf(s);
				break;
			
			//-- robot examined params
			case 8:
				parameters.odomVarianceCoeff = Double.valueOf(s);
				break;
			case 9:
				parameters.communicationRadius = Integer.valueOf(s);
				break;
			
			//-- optimised params
			case 10:
				parameters.timeToRandWalk = Integer.valueOf(s);
				break;
			case 11:
				parameters.timeToWalkNearBase = Integer.valueOf(s);
				break;
			
			case 12:
				parameters.neighbourhoodSearchRadius = Integer.valueOf(s);
				break;
			case 13:
				parameters.neighbourhoodSearchTime = Integer.valueOf(s);
				break;
			case 14:
				parameters.nearDistanceToBase = Integer.valueOf(s);
				break;
			
			}
			counter++;
			speed = 1;
			startSimulation = true;
			parameters.numOfRuns = 20;
			
		}
		settings.setIsConsoleOnlyBuild(consoleOnly);
		new RobotForaging(startSimulation, speed);
		
				
				
	}
        
            
    

}
