package net.lenkaspace.robotForaging.model;

import java.util.ArrayList;

import net.lenkaspace.creeper.helpers.CRMaths;
import net.lenkaspace.creeper.model.CRBaseDynamicModel;
import net.lenkaspace.creeper.model.CRBaseSituatedModel;
import net.lenkaspace.creeper.model.CRSettings;
import net.lenkaspace.creeper.model.CRWorld;
import net.lenkaspace.creeper.report.CRReportController;
import net.lenkaspace.creeper.report.CRTimeSeriesReport;
import net.lenkaspace.creeper.view.CRRenderer;
import net.lenkaspace.creeper.vo.CRVector3d;
 

public class Robot extends CRBaseDynamicModel {
	
	private static int RANDOM_WALK_TURN_FREQUENCY_MIN = 1; // in seconds
	private static int RANDOM_WALK_TURN_FREQUENCY_MAX = 2;
	private static int TURN_ANGLE_MIN = 5; //5
	private static int TURN_ANGLE_MAX = 45; //45
	
	private static int OBSTACLE_DETECTION_RADIUS = 30;
	
	private static double BACKING_UP_TIME = CRSettings.getSingleton().getMaxUpdateDelay()*0.25;
	private static double UNLOAD_PER_FRAME = 0.01; //unloaded in 2 seconds
	private static double LOAD_PER_FRAME = 0.004; //loaded in 5 seconds 0.004
	
	private static double DEPOSIT_ESTIMATION_VARIANCE = 0.05;
	
	public static double ENERGY_NAVIGATE_TO_DEPOSIT = 0.00001;
	public static double ENERGY_NAVIGATE_TO_BASE = 0.00005;
	public static double ENERGY_LOAD_RESOURCE = 0.00005;
	public static double ENERGY_UNLOAD_RESOURCE = 0.00002;
	
	private static enum STATE {
		RANDOM_WALK, NAVIGATE_TO_BASE, NAVIGATE_TO_DEPOSIT, LOAD_RESOURCE, UNLOAD_RESOURCE, SEARCH_NEIGHBOURHOOD, RANDOM_WALK_NEAR_BASE
	}
	
	private class CommunicationSignal {
		public CRVector3d depositPosition = null;
		public double depositEnergyEfficiency = -1;
		public int signalerId = -1;
	}
	
	private class SensoryInfo {
		public CRBaseSituatedModel objectNear = null;
		public CRBaseSituatedModel objectCollided = null;		
		public boolean isAtBase = false;
		public ArrayList<CommunicationSignal> receivedSignals;
		
		public SensoryInfo() {
			objectNear = null;
			objectCollided = null;
			isAtBase = false;
			receivedSignals = new ArrayList<CommunicationSignal>();
		}
	}
	
	private class MotorCommand {
		public int rotationChange = 0;
		public double newThrust = 1;
		
		public MotorCommand(int rotationChange_, double thrust_) {
			rotationChange = rotationChange_;
			newThrust = thrust_;
		}
		
		public String toString() {
			return new String("thrust: " + newThrust + " rot change:" + rotationChange);
		}
	}
	
	
	
	private SensorArea collisionDetectionArea;
	private SensorArea odometryLocationDummy;
	private SensorArea depoLocationDummy;
	
	private double backUpTime;
	private STATE state;
	private double nextRandomTurnTime;
	private double carriedResourceGross;
	private double carriedResourceNet;
	private CRVector3d depositPosition;
	private double depositEnergyEfficiency;
	
	private CRVector3d odomPosition;
	private double odomRotation;
	private boolean odomShouldRecord;
	
	private double searchTimeLeft;
	private double randWalkTimeLeft;
	private double walkNearBaseTimeLeft;
	private boolean objectNearIsInFront;
	
	private boolean depositSwitched;
	private double depositId;
	private double distanceTraveled;
	private int recruiterId;
	private int resourceRecruiterId;
	private double worstFoundQuality;
	private boolean shouldReturnToDoSearch;
	
	//-- array that is a sequence of when depos were found according to:
	//						|	Better than base	|	Not better than base quality	|
	//	-----------------------------------------------------------------------------------
	//	By Neighb Search	|		SB				|		SN
	//	During going to depo|		DB				|		DN
	//	Random search		|		RB				|		RN
	
	private ArrayList<String> foundDepoEvents;
	private boolean foundDepoEventProcessed;
	private String foundDepoEventType;
	
	public Robot(int id_, CRVector3d position_, double rotation_, CRWorld world_) {
		super(id_, position_, new CRVector3d(10,10,0), rotation_, SHAPE.CIRCLE, "robot.png");
		world = world_;
		
		collisionDetectionArea = new SensorArea(id, position, new CRVector3d(OBSTACLE_DETECTION_RADIUS*2, OBSTACLE_DETECTION_RADIUS*2,0), CRRenderer.CR_BLANK_CIRCLE);
		world.addSituatedModel(collisionDetectionArea);
		
		odometryLocationDummy = new SensorArea(id, position, new CRVector3d(5,5,0), CRRenderer.CR_RED_CIRCLE);
		world.addSituatedModel(odometryLocationDummy);
		
		depoLocationDummy = new SensorArea(id, position, new CRVector3d(5,5,0), CRRenderer.CR_GRAY_DOT);
		world.addSituatedModel(depoLocationDummy);
		
		foundDepoEvents = new ArrayList<String>();
		
		depositPosition = null;
		searchTimeLeft = 0;
		
		carriedResourceGross = 0;
		nextRandomTurnTime = getNextRandomTurnTime();
		backUpTime = -1;
		depositEnergyEfficiency = -1;
		depositSwitched = false;
		depositId = -1;
		distanceTraveled = 0;
		
		worstFoundQuality = -1;
		foundDepoEventProcessed = false;
		foundDepoEventType = "";
		
		this.setState(STATE.RANDOM_WALK);
		resetOdometry();
		this.toolTip = "" + id;
	}
	
	public void update() {
		
		//-- move the detection areas with the body
		collisionDetectionArea.setRotation(rotation);
		collisionDetectionArea.setPosition(position);
		collisionDetectionArea.setIsVisible(Settings.getSingleton().showAvoidRadius);
		
		Parameters parameters = Parameters.getSingleton();
		CRReportController reportController = world.getController().getReportController();
		
		setImageFileName("robot.png");
		if (state == STATE.SEARCH_NEIGHBOURHOOD) {
			setImageFileName("robotBlue.png");
		} else if (state == STATE.LOAD_RESOURCE || state == STATE.UNLOAD_RESOURCE) {
			setImageFileName("robotBlack.png");
		} else if (state == STATE.NAVIGATE_TO_BASE) {
			setImageFileName("robotCarying.png");
		} else if (state == STATE.RANDOM_WALK_NEAR_BASE) {
			setImageFileName("robotGreen.png");
		} else if (state == STATE.NAVIGATE_TO_DEPOSIT) {
			setImageFileName("robotLightBlue.png");
		}
		
		//-- get sensory info
		SensoryInfo sensoryInfo = getSensoryInfo();
	
		//-- get motor commands from layers
		selectResource(sensoryInfo);
		MotorCommand randWalkMotorComm = doRandomWalk(sensoryInfo);
		MotorCommand navigateToDestinationMotorComm = navigateToDestination(sensoryInfo);
		MotorCommand processCurrentPosition = processCurrentPosition(sensoryInfo);
		MotorCommand avoidMotorComm = avoidObstacles(sensoryInfo);
		
		//-- smoothen the carried resource amount
		if (carriedResourceGross > 1) { carriedResourceGross = 1; }
		if (carriedResourceGross < 0) { carriedResourceGross = 0; }
		
		//-- do the motor commands
		MotorCommand motorCommToUse = new MotorCommand(0,0);
		if (avoidMotorComm != null) {
			motorCommToUse = avoidMotorComm;
			this.setImageFileName("robotRed.png");
		} else if (processCurrentPosition != null) {
			motorCommToUse = processCurrentPosition;
		} else if (navigateToDestinationMotorComm != null) {
			motorCommToUse = navigateToDestinationMotorComm;
		} else if (randWalkMotorComm != null) {
			motorCommToUse = randWalkMotorComm;
		}
		
		//System.out.println(state.toString() + "   " + carriedResourceGross + "   " + carriedResourceNet);
		
		//-- rotation boundaries
		if (motorCommToUse.rotationChange > TURN_ANGLE_MAX) { motorCommToUse.rotationChange = TURN_ANGLE_MAX; }
		else if (motorCommToUse.rotationChange < -TURN_ANGLE_MAX) { motorCommToUse.rotationChange = -TURN_ANGLE_MAX; }
		
		//-- calculate odometry
		double measuredDist = motorCommToUse.newThrust;
		odomRotation += motorCommToUse.rotationChange;
		if (odomRotation > 180){ odomRotation = -360 + odomRotation; }
		else if (odomRotation < -180){ odomRotation = 360 + odomRotation; }
    	
		odomShouldRecord = true;
		if (sensoryInfo.objectCollided != null && sensoryInfo.objectCollided.getClass() != Base.class) { odomShouldRecord = false; }
		if (motorCommToUse.newThrust == 0) { odomShouldRecord = false; }
		
		if (odomShouldRecord) {
	    	odomPosition.x += Math.sin(Math.toRadians(odomRotation)) * measuredDist;
	    	odomPosition.y += -Math.cos(Math.toRadians(odomRotation)) * measuredDist;
	    	if (odomPosition.x - size.x/4 > world.getSize().x ) { odomPosition.x = - size.x / 4;	} 
			else if (odomPosition.x + size.x/4 < 0 ) { odomPosition.x = world.getSize().x + size.x / 4; }
			if (odomPosition.y - size.y/4 > world.getSize().y ) { odomPosition.y = - size.y / 4; }
			else if (odomPosition.y + size.y /4 < 0 ) { odomPosition.y = world.getSize().y  + size.y / 4; }
		}
		
		//-- perform motor command, with added noize
		double positionErrorVariance = parameters.odomVarianceCoeff*Math.abs(motorCommToUse.newThrust);
		double rotationErrorVariance = parameters.odomVarianceCoeff*Math.abs(motorCommToUse.rotationChange);
		double positionError = CRMaths.getRandomGaussian(0, positionErrorVariance);
		double rotationError = CRMaths.getRandomGaussian(0, rotationErrorVariance);
		//System.out.println("ERROR POS " + positionError + "  ROT " + rotationError + " req rot " + motorCommToUse.rotationChange + "  err var " + rotationErrorVariance);
		this.setRotation(rotation + motorCommToUse.rotationChange + rotationError);
		thrustForce = motorCommToUse.newThrust + positionError ;
		
		//-- move dummy
		if (Settings.getSingleton().showOdometryPosition) {
			CRVector3d dummyPos = new CRVector3d(odomPosition);
			odometryLocationDummy.setPosition(dummyPos);
			odometryLocationDummy.setIsVisible(true);
		} else {
			odometryLocationDummy.setIsVisible(false);
		}
		
		if (depositPosition != null && Settings.getSingleton().showEstimDepoPosition) {
			depoLocationDummy.setIsVisible(true);
			depoLocationDummy.setPosition(depositPosition);
		} else {
			depoLocationDummy.setIsVisible(false);
		}
		
		//-- move
		super.update();	
		distanceTraveled += Math.abs(thrustForce);
		
		//-- handle collision
		ArrayList<CRBaseSituatedModel> worldObjects = ((World)world).getRelevantObjectsObject(this);
		for (CRBaseSituatedModel object : worldObjects) {
			Class<? extends CRBaseSituatedModel> objClass = object.getClass();
			if (object != this && objClass != SensorArea.class && objClass != Base.class && objClass != Deposit.class ) {//!(objClass==Deposit.class && state == STATE.LOAD_RESOURCE)
				//-- test for collision
				if (this.isCollidingWith(object)) {
					CRVector3d relVector = this.getRelativeVectorTo(object.getPosition());
					if (relVector.y * thrustForce >= 0)  {  //Math.abs(relVector.x) <= 5 && 
						//isColliding = true;
						if (previousPosition != null && previousPosition.isValid()) {
							setPosition(previousPosition);
						}
						break;
					}
				}	
			}
		}

		//-- write into reports
		CRTimeSeriesReport robotReport = (CRTimeSeriesReport)reportController.getReport(Settings.REPORT_ROBOT + id);
		if (robotReport != null) {
			robotReport.addValue(Math.abs(CRMaths.getDistanceOfPoints(odomPosition.x, odomPosition.y, position.x, position.y)), Settings.VAR_ODOM_ERROR);
		} else {
			System.out.println("! Couldn't get robot report " + id);
		}
		
		//System.out.println(id + " update end");
	}
	
	
	/**
	 * Select resource layer. Doesn't affect motors but saves stuff into memory.
	 * @param sensoryInfo_ SensoryInfo
	 */
	private void selectResource(SensoryInfo sensoryInfo_) {
		//-- if senses a resource nearby, get a vector towards it and set it as destination
		if (sensoryInfo_.objectNear != null) {
			if (sensoryInfo_.objectNear.getClass() == Deposit.class && state != STATE.NAVIGATE_TO_BASE) {
				//-- assume robot can measure relative vector to the object. 
				//   Then, based on its own position, it can calculate position of the object and the 
				//	 error of that position will be = error of odometry	
				CRVector3d realDepoPos = sensoryInfo_.objectNear.getPosition();
				CRVector3d estimDepoPos = new CRVector3d(realDepoPos);
				estimDepoPos.x += odomPosition.x - position.x;
				estimDepoPos.y += odomPosition.y - position.y;
				
				//-- only set this as a new target if it is safe to assume it's new, i.e. either no previous target or 
				//   position outside of original detection area
				boolean setNewTarget = false;
				if (depositPosition == null) {
					setNewTarget = true;
				} else {
					double d = CRMaths.getDistanceOfPoints(depositPosition.x, depositPosition.y, estimDepoPos.x, estimDepoPos.y);
					if (d > OBSTACLE_DETECTION_RADIUS*0.1) {
						setNewTarget = true;
					}
				}
				if (setNewTarget) {
					//-- only set if better target
					Deposit deposit = (Deposit)sensoryInfo_.objectNear;
					double quantityVariance = DEPOSIT_ESTIMATION_VARIANCE*deposit.getQuantity();
					double qualityVariance = DEPOSIT_ESTIMATION_VARIANCE*deposit.getResourceConcentration();
					double energyEfficiency = estimateEnergyEfficiency(estimDepoPos,CRMaths.getRandomGaussian(deposit.getQuantity(), quantityVariance), CRMaths.getRandomGaussian(deposit.getResourceConcentration(), qualityVariance));
					//System.out.println("OLD " + depositEnergyEfficiency + "  NEW" + energyEfficiency);
					if (energyEfficiency > depositEnergyEfficiency) {
						//-- find out if this deposit was found by neighborhood search
						if (depositId >= 0 && depositId != deposit.getId() && state == STATE.SEARCH_NEIGHBOURHOOD) {
							depositSwitched = true;
							//System.out.println("SWITCH");
						}
						depositId = deposit.getId();
						depositEnergyEfficiency = energyEfficiency;
						depositPosition = new CRVector3d(estimDepoPos);
						
						//-- remember how it was found
						if (foundDepoEventType.length() == 0) {
							if (state == STATE.SEARCH_NEIGHBOURHOOD) {
								foundDepoEventType = "S";
							} else if (state == STATE.NAVIGATE_TO_DEPOSIT) {
								foundDepoEventType = "D";
							} else if (state == STATE.RANDOM_WALK || state == STATE.RANDOM_WALK_NEAR_BASE) {
								foundDepoEventType = "R";
							}	
						}
						setState(STATE.NAVIGATE_TO_DEPOSIT);	
					}
					
				}
			}
		}
		
		//-- set new target based on received signals
		if (Parameters.getSingleton().useBSwarm) {
			for (CommunicationSignal signal : sensoryInfo_.receivedSignals) {
				//System.out.println(id + "   " + signal.depositEnergyEfficiency + "    " + depositEnergyEfficiency);
				if (signal.depositEnergyEfficiency > depositEnergyEfficiency) {
					//-- only set this as a new target if it is safe to assume it's new, i.e. either no previous target or 
					//   position outside of original detection area
					boolean setNewTarget = false;
					if (depositPosition == null) {
						setNewTarget = true;
					} else {
						double d = CRMaths.getDistanceOfPoints(depositPosition.x, depositPosition.y, signal.depositPosition.x, signal.depositPosition.y);
						if (d > OBSTACLE_DETECTION_RADIUS) {
							setNewTarget = true;
						}
					}
					if (setNewTarget) {
						depositEnergyEfficiency = signal.depositEnergyEfficiency;
						depositPosition = new CRVector3d(signal.depositPosition);
						recruiterId = signal.signalerId;
						//System.out.println(id + " getting signal from " + recruiterId);
						//-- remember how it was found
						if (foundDepoEventType.length() == 0) {
							if (state == STATE.RANDOM_WALK_NEAR_BASE) {
								foundDepoEventType = "B";
							} else {
								foundDepoEventType = "C";
							}
						}
						//System.out.println("SWITHCING TO SIGNALED DEPO");
						//-- go there as well, if currently not busy 
						if (state != STATE.LOAD_RESOURCE && state != STATE.UNLOAD_RESOURCE && state != STATE.NAVIGATE_TO_BASE) {
							setState(STATE.NAVIGATE_TO_DEPOSIT);
							//System.out.println(">> GOING");
						}
					}
				}
			}
		}
	}
	
	/**
	 * Random walk layer. Turn at random intervals by a random angle.
	 * Also turn when doing random neighbourhood search and outside of area
	 * @param sensoryInfo_ SensoryInfo
	 * @return MotorCommand
	 */
	private MotorCommand doRandomWalk(SensoryInfo sensoryInfo_) {
		MotorCommand comm = new MotorCommand(0,1);
		int currentTime = world.getController().getTimeCounterSinceTrialStart();
		Parameters parameters = Parameters.getSingleton();
		
		//-- random turn if it is time to turn or if doing neighbourhood search and outside of area
		if (currentTime >= nextRandomTurnTime) {
			comm.rotationChange = CRMaths.getRandomInteger(TURN_ANGLE_MIN, TURN_ANGLE_MAX, true);
			nextRandomTurnTime = getNextRandomTurnTime();
		} else {
			//-- NEIGHBORHOOD SEARCH
			if (searchTimeLeft > 0) {
				searchTimeLeft--;
				//-- does it believe it's outside of the neighborhood?
				if (CRMaths.getDistanceOfPoints(odomPosition.x, odomPosition.y, depositPosition.x, depositPosition.y) > parameters.neighbourhoodSearchRadius - size.x) {
					comm.rotationChange = getRotationCommandTowardsPosition(depositPosition);
					nextRandomTurnTime = getNextRandomTurnTime();
				}
				if (searchTimeLeft == 0) {
					//-- end of search, resume random walk and forget the resource location
					setState(STATE.RANDOM_WALK);
					//depositPosition = null;
					depositEnergyEfficiency = -1;
				}
			}
			//-- RANDOM WALK NEAR BASE
			if (state == STATE.RANDOM_WALK_NEAR_BASE) {
				CRVector3d basePos = ((World)world).getBase().getPosition();
				if (CRMaths.getDistanceOfPoints(odomPosition.x, odomPosition.y, basePos.x, basePos.y) > parameters.nearDistanceToBase - size.x) {
					comm.rotationChange = getRotationChangeTowardsPosition(basePos);
					nextRandomTurnTime = getNextRandomTurnTime();
				}
				//-- should end?
				if (walkNearBaseTimeLeft >= 0) {
					if (walkNearBaseTimeLeft == 0) {
						setState(STATE.RANDOM_WALK);
					}
					walkNearBaseTimeLeft--;
				}
			}
		}
		
		//-- check if random walk should end and should return to near base:
		if (Parameters.getSingleton().useBSwarm) {
			if (state == STATE.RANDOM_WALK && randWalkTimeLeft >= 0) {
				if (randWalkTimeLeft == 0) {
					setState(STATE.RANDOM_WALK_NEAR_BASE);
				}
				randWalkTimeLeft--;
			}
		}
		return comm;
		
	}
	
	
	/**
	 * Navigate to a destination layer. Sets rotation to reach a destination in memory
	 * @param sensoryInfo_ SensoryInfo
	 * @return MotorCommand
	 */
	private MotorCommand navigateToDestination(SensoryInfo sensoryInfo_) {
		MotorCommand comm = null;
		if (state == STATE.NAVIGATE_TO_BASE) {	
			CRVector3d basePos = ((World)world).getBase().getPosition();
			comm = new MotorCommand(getRotationChangeTowardsPosition(basePos),0.5);	
			//System.out.println("NAV TO BASE " + carriedResource);
		} else if (state == STATE.NAVIGATE_TO_DEPOSIT) {
			if (depositPosition != null) {
				//-- assume only using odometry to navigate to depos
				comm = new MotorCommand(getRotationCommandTowardsPosition(depositPosition),1);	
				if (Math.abs(comm.rotationChange) <= 7) {
					comm.rotationChange = 0;
				}
				//System.out.println("NAV TO RES");
			}
		}
		return comm;
	}
	
	/**
	 * Process what robot is currently colliding with or the position it is at.
	 * @param sensoryInfo_ SensoryInfo
	 * @return MotorCommand
	 */
	private MotorCommand processCurrentPosition(SensoryInfo sensoryInfo_) {
		MotorCommand comm = null;
		Parameters parameters = Parameters.getSingleton();
		if (sensoryInfo_.objectCollided != null) {
			Class<? extends CRBaseSituatedModel> objCollidedClass = sensoryInfo_.objectCollided.getClass();
			if (objCollidedClass == Deposit.class) {

				//-- extract from the resource if not full, don't go anywhere
				if (carriedResourceGross < 1) {
					//-- capture the even when just started gathering
					if (carriedResourceGross == 0 && !foundDepoEventProcessed) {
						foundDepoEventProcessed = true;
						Deposit deposit = (Deposit)sensoryInfo_.objectCollided;
						depositId = deposit.getId();
						double repActualDepositEnergyEfficiency = estimateEnergyEfficiency(deposit.getPosition(),deposit.getQuantity(), deposit.getResourceConcentration());
						double repDepositEnergyEfficiency = depositEnergyEfficiency;
						
						//System.out.println(repActualDepositEnergyEfficiency);
						
						CRTimeSeriesReport robotReport = (CRTimeSeriesReport)world.getController().getReportController().getReport(Settings.REPORT_ROBOT + id);
						robotReport.addValue(deposit.getResourceConcentration(), Settings.VAR_DEPO_QUALITY);
						robotReport.addValue(repActualDepositEnergyEfficiency, Settings.VAR_DEPO_EE);
						robotReport.addValue(repDepositEnergyEfficiency, Settings.VAR_DEPO_ESTIM_EE);
						if (depositSwitched) {
							//-- this depo was found during neighborhood, search, record the time
							double searchTimeSpent = parameters.neighbourhoodSearchTime*CRSettings.getSingleton().getMaxUpdateDelay() - searchTimeLeft;
							robotReport.addValue(searchTimeSpent, Settings.VAR_SEARCH_TIME);
							robotReport.addValue(searchTimeSpent/(parameters.neighbourhoodSearchTime*CRSettings.getSingleton().getMaxUpdateDelay()), Settings.VAR_SEARCH_TIME_PERCENT);
						//	System.out.println("NOW " + searchTimeLeft +  "    " + searchTimeSpent + " = " + (searchTimeSpent/(parameters.neighbourhoodSearchTime*CRSettings.getSingleton().getMaxUpdateDelay()))  );
						}
						robotReport.addValue(recruiterId, Settings.VAR_RECRUITER_ID);
						//-- report on distances
						Base base = ((World)world).getBase();
						CRVector3d basePos = base.getPosition();
						robotReport.addValue(CRMaths.getDistanceOfPoints(position.x, position.y, basePos.x, basePos.y) - (base.getSize().x / 2.0 + size.x/2.0), Settings.VAR_DISTANCE_FROM_BASE);
						robotReport.addValue(distanceTraveled, Settings.VAR_DISTANCE_TRAVELLED); 
						
						//-- report on the type of event
						if (foundDepoEventType.length() == 0) {
							if (state == STATE.SEARCH_NEIGHBOURHOOD) {
								foundDepoEventType = "S";
							} else if (state == STATE.NAVIGATE_TO_DEPOSIT) {
								foundDepoEventType = "D";
							} else if (state == STATE.RANDOM_WALK || state == STATE.RANDOM_WALK_NEAR_BASE) {
								foundDepoEventType = "R";
							}
						}
						if (deposit.getResourceConcentration() > ((World)world).getBaseDepoQuality()) {
							foundDepoEventType += "B";
						} else {
							foundDepoEventType += "N";
						}
						
						//System.out.println(id + "  " + deposit.getResourceConcentration() + "  " + foundDepoEventType + "  V=" + deposit.getQuantity() + "  recruiter" + recruiterId);
						foundDepoEvents.add(new String(foundDepoEventType));	
						foundDepoEventType = "";
								
						//--
						depositSwitched = false;
						resourceRecruiterId = recruiterId; // remember who recruited for this depo, because recruiter id can change during trip to base
						//-- remember worst quality resource found
						//   also, do not return here to do neighborhood search if quality is worst than worst encountered
						if (deposit.getResourceConcentration() <= worstFoundQuality) {
							shouldReturnToDoSearch = false;
						} else {
							shouldReturnToDoSearch = true;
						}
						if (deposit.getResourceConcentration() < worstFoundQuality || worstFoundQuality == -1) {
							worstFoundQuality = deposit.getResourceConcentration();
						} 
						//System.out.println(">>>> EVENT ee" + repActualDepositEnergyEfficiency + " estim ee " + repDepositEnergyEfficiency + "  tSearch" + searchTimeLeft + "   D from base " + (CRMaths.getDistanceOfPoints(position.x, position.y, basePos.x, basePos.y) - (base.getSize().x / 2.0 + size.x/2.0)) + "  Travelled " + distanceTraveled );
					}
					comm = new MotorCommand(0,0);
					setState(STATE.LOAD_RESOURCE);
					Deposit.Packet packet = ((Deposit)sensoryInfo_.objectCollided).extractResource(LOAD_PER_FRAME);
					carriedResourceGross += packet.grossAmount;
					carriedResourceNet += packet.netAmount;
					searchTimeLeft = 0;
				}
				
				//-- if there is nothing to extract anymore, return to base and remember this position
				if (state == STATE.LOAD_RESOURCE) {
					if (((Deposit)sensoryInfo_.objectCollided).getQuantity() <= 0 || carriedResourceGross == 1) {
						setState(STATE.NAVIGATE_TO_BASE);
						foundDepoEventProcessed = false;
						depositPosition = new CRVector3d(odomPosition);
						Deposit deposit = (Deposit)sensoryInfo_.objectCollided;
						depositEnergyEfficiency = estimateEnergyEfficiency(odomPosition,CRMaths.getRandomGaussian(deposit.getQuantity(), DEPOSIT_ESTIMATION_VARIANCE), CRMaths.getRandomGaussian(deposit.getResourceConcentration(), DEPOSIT_ESTIMATION_VARIANCE));
						
						
					}
				}
			}
		} else {
			//-- test if reached destination
			if (depositPosition != null && state == STATE.NAVIGATE_TO_DEPOSIT) {
				boolean isAtDepo = CRMaths.getDistanceOfPoints(odomPosition.x, odomPosition.y, depositPosition.x, depositPosition.y) < size.x/2;
				
				if (isAtDepo && searchTimeLeft <= 0) { //isAt(depositPosition,20)
					searchTimeLeft = parameters.neighbourhoodSearchTime * CRSettings.getSingleton().getMaxUpdateDelay();
					setState(STATE.SEARCH_NEIGHBOURHOOD);
					//depositEnergyEfficiency = -1;
					
					//depositPosition = null;
				}
			}
		}
		
		//-- test if reached base
		if (sensoryInfo_.isAtBase) {
			if (carriedResourceGross > 0) {
				//-- not going anywhere, unload
				setState(STATE.UNLOAD_RESOURCE);
				carriedResourceGross -= UNLOAD_PER_FRAME;
				comm = new MotorCommand(0,0);
			} else {
				if (state == STATE.UNLOAD_RESOURCE) {
					//-- event: just finished unloading, register this in base
					carriedResourceGross = 0;
					((World)world).getBase().addResource(carriedResourceNet, resourceRecruiterId);
					carriedResourceNet = 0;
					if (parameters.useNeighborhoodSearch && shouldReturnToDoSearch) {
						setState(STATE.NAVIGATE_TO_DEPOSIT);
					} else {
						setState(STATE.RANDOM_WALK);
					}
				}
				
			}
			resetOdometry();
			distanceTraveled = 0;
		}
		return comm;
	}
	
	/**
	 * Avoid obstacles layer. Can subsume all layers by going backwards and then randomly turning.
	 * @param sensoryInfo_ SensoryInfo
	 * @return MotorCommand
	 */
	private MotorCommand avoidObstacles(SensoryInfo sensoryInfo_) {
		MotorCommand comm = null;
		
		if (sensoryInfo_.objectNear != null) {
			//-- test if object is in front to start avoiding and if it's of correct type
			CRVector3d relVector = this.getRelativeVectorTo(sensoryInfo_.objectNear.getPosition());
			double xDistanceInLocal = Math.sin(Math.abs(relVector.x)) * Math.abs(relVector.y);
			Class<? extends CRBaseSituatedModel> objClass = null;
			if ((xDistanceInLocal < size.x && relVector.y * thrustForce >= 0)) { //45deg <=> relVector.x = 1
				objClass = sensoryInfo_.objectNear.getClass();
			} else if (sensoryInfo_.objectCollided != null) {
				objClass = sensoryInfo_.objectCollided.getClass();
			}
			if (objClass != null) {
				
				if ( (objClass != Deposit.class) || (objClass == Deposit.class && state == STATE.NAVIGATE_TO_BASE) ) {
					if (backUpTime < 0) {
						backUpTime = BACKING_UP_TIME;
						if (relVector.y > 0) {
							objectNearIsInFront = true;
						} else {
							objectNearIsInFront = false;
						}
					}
				}
			}
		}
		
		if (backUpTime > 0) {
			backUpTime--;
			comm = new MotorCommand(0,-1);
			if (!objectNearIsInFront) {
				comm.newThrust = 1;
			}
			if (backUpTime <= 0) {
				comm.rotationChange = CRMaths.getRandomInteger(TURN_ANGLE_MIN, TURN_ANGLE_MAX, true);
			}
		} 
		return comm;
		
	}
		
	//============================== SIGNALS ========================
	
	/**
	 * Signal about found resource and its estimate location and energy efficiency
	 * @return CommunicationSignal signal
	 */
	public CommunicationSignal getCommunicationSignal() {
		if (depositPosition != null) {
			CommunicationSignal signal = new CommunicationSignal();
			signal.depositPosition = new CRVector3d(depositPosition);
			signal.depositEnergyEfficiency = depositEnergyEfficiency;
			//-- decide who is the robot originaly signaling about this
			if (recruiterId < 0) {
				signal.signalerId = this.id;
			} else {
				signal.signalerId = recruiterId;
			}
			return signal;
		}
		return null;
	}
	
	//============================== SENSORS ========================
	
	/**
	 * Ask sensor what's in front etc.
	 * @return SensoryInfo
	 */
	private SensoryInfo getSensoryInfo() {
		SensoryInfo info = new SensoryInfo();
		ArrayList<CRBaseSituatedModel> worldObjects = ((World)world).getRelevantObjectsObject(this);
		for (CRBaseSituatedModel object : worldObjects) {
			Class<? extends CRBaseSituatedModel> objClass = object.getClass();
			if (object != this && objClass != SensorArea.class && objClass != Base.class) {
				
				//-- test if infra sensor detecting
				if (collisionDetectionArea.isCollidingWith(object)) {
					info.objectNear = object;			
				}
				
				//-- test for collision
				if (this.isCollidingWith(object)) {
					CRVector3d relVector = this.getRelativeVectorTo(object.getPosition());
					if (relVector.y * thrustForce >= 0)  {  //Math.abs(relVector.x) <= 5 && 
						info.objectNear = object;
						info.objectCollided = object;
					}
				}	
			}
			
			//-- test for signals from other robots
			if (object != this && objClass == Robot.class) {
				Parameters params = Parameters.getSingleton();
				if (CRMaths.getDistanceOfPoints(position.x, position.y, object.getPosition().x, object.getPosition().y) <= params.communicationRadius) {
					CommunicationSignal signal = ((Robot)object).getCommunicationSignal();
					if (signal != null) {
						//System.out.println(id + " received signal from" + object.getId());
						info.receivedSignals.add(signal);
					}
				}
			}
		}
		//-- test for collision with base
		if (this.isCollidingWith(((World)world).getBase())) {
			info.isAtBase = true;
		}
		return info;
	}
	
	//============================== HELPERS ========================
	/**
	 * Get rotation change based on where robot believes it is
	 * @param position_ CRVector3d position to rotate towards
	 * @return int rotation change
	 */
	private int getRotationCommandTowardsPosition(CRVector3d position_) {
		int retRot = 0;
		CRVector3d temp = new CRVector3d(position);
		if (odomPosition != null) {
			position.copyFrom(odomPosition);
		}
		retRot = getRotationChangeTowardsPosition(position_) + CRMaths.getRandomInteger(0, TURN_ANGLE_MAX, true);
		position.copyFrom(temp);
		return retRot;
	}
	/**
	 * Setting a robot state
	 * @param state_
	 */
	private void setState(STATE state_) {
		state = state_;	
		if (state == STATE.RANDOM_WALK) {
			randWalkTimeLeft = Parameters.getSingleton().timeToRandWalk * CRSettings.getSingleton().getMaxUpdateDelay();
			//-- any time it had to start random walk, any recruitment info was bad
			recruiterId = -1;
			foundDepoEventType = "";
		} else if (state == STATE.RANDOM_WALK_NEAR_BASE) {
			walkNearBaseTimeLeft = Parameters.getSingleton().timeToWalkNearBase * CRSettings.getSingleton().getMaxUpdateDelay();
		}
	}
	
	private double estimateEnergyEfficiency(CRVector3d estimatedPosition_, double estimatedQuantity_, double estimatedQuality_) {
		CRVector3d basePos = ((World)world).getBase().getPosition();
		double dToBase = Math.abs(CRMaths.getDistanceOfPoints(basePos.x, basePos.y, estimatedPosition_.x, estimatedPosition_.y));
		//-- estimate energy efficiency, assume robot can find out how much resource and of what quality is there
		double grossGainedIncome = estimatedQuantity_* estimatedQuality_;
		double energySpent = dToBase*ENERGY_NAVIGATE_TO_DEPOSIT + dToBase*ENERGY_NAVIGATE_TO_BASE + (1/LOAD_PER_FRAME)*ENERGY_LOAD_RESOURCE + (1/UNLOAD_PER_FRAME)*ENERGY_UNLOAD_RESOURCE;
		double energyEfficiency = (grossGainedIncome-energySpent)/energySpent; 
		return energyEfficiency;
	}
	
	public static double calculateEnergyEfficiency(double dToBase_, double estimatedQuantity_, double estimatedQuality_) {
		double grossGainedIncome = estimatedQuantity_* estimatedQuality_;
		double energySpent = dToBase_*ENERGY_NAVIGATE_TO_DEPOSIT + dToBase_*ENERGY_NAVIGATE_TO_BASE + (1/LOAD_PER_FRAME)*ENERGY_LOAD_RESOURCE + (1/UNLOAD_PER_FRAME)*ENERGY_UNLOAD_RESOURCE;
		double energyEfficiency = (grossGainedIncome-energySpent)/energySpent; 
		return energyEfficiency;
	}
	
	/**
	 * Assume robot knows where exactly within the base it is and what is its rotation
	 */
	private void resetOdometry() {
		odomPosition = new CRVector3d(position);
		odomRotation = rotation;
	}
	
	private double getNextRandomTurnTime() {
		double secondTicks = CRSettings.getSingleton().getMaxUpdateDelay();
		double range = RANDOM_WALK_TURN_FREQUENCY_MAX*secondTicks - RANDOM_WALK_TURN_FREQUENCY_MIN*secondTicks;
		return world.getController().getTimeCounterSinceTrialStart() + RANDOM_WALK_TURN_FREQUENCY_MIN*secondTicks + range*CRMaths.getRandomDouble(true);
	}
	
	private int getRotationChangeTowardsPosition(CRVector3d position_) {
		CRVector3d relVector = this.getRelativeVectorTo(position_);
		int rotationChange = -(int) Math.toDegrees(relVector.x);
		if (relVector.y*thrustForce <= 0) { //Math.abs(rotationChange) < TURN_ANGLE_MAX
			rotationChange = TURN_ANGLE_MAX;
		}
		return rotationChange;
	}
	
	//---- getters / setters
	public ArrayList<String> getFoundDepoEvents() { return foundDepoEvents;	}
	
}
