package net.lenkaspace.robotForaging.model;

import java.util.ArrayList;

import net.lenkaspace.creeper.CRController;
import net.lenkaspace.creeper.helpers.CRFileIOHelper;
import net.lenkaspace.creeper.helpers.CRMaths;
import net.lenkaspace.creeper.model.CRBaseSituatedModel;
import net.lenkaspace.creeper.model.CRBinWorld;
import net.lenkaspace.creeper.view.CRRenderer;
import net.lenkaspace.creeper.vo.CRVector3d;

public class World extends CRBinWorld {

	public static String TEST = "Test";
	public static String RANDOM = "Random";
	public static String RANDOM_DISTANT = "Random distant deposits";
	public static String PATCHY = "Patchy";

	private static int MAX_DEPOSITS = 200;
	private static int MAX_TOTAL_NET_AMOUNT = 100;

	private Base base;
	private ArrayList<Deposit> deposits;
	private ArrayList<Robot> robots;
	private double baseDepoQuality;
	private double newRobotPositionExtraDistance;

	public World(CRController controller_) {
		super(new CRVector3d(400, 400, 0), 0, new CRVector3d(4000, 4000, 0),
				controller_);
		deposits = new ArrayList<Deposit>();
		robots = new ArrayList<Robot>();

		// -- add to scenarios
		//this.getScenarios().add(TEST);
		this.getScenarios().add(RANDOM);
		this.getScenarios().add(RANDOM_DISTANT);
		this.getScenarios().add(PATCHY);

	}
	

	public void onTrialStart(int trialNumber_, int runNumber_) {
		
		super.onTrialStart(trialNumber_, runNumber_);
		this.clearChildren();

		//-- create base
		base = new Base(new CRVector3d(2000,2000,0));
		this.addDynamicModel(base);
		Parameters parameters = Parameters.getSingleton();
		int i;
		if (parameters.scenario != TEST) {
			String info = "X\tY\tOrigId\tQ\tC\tdOrig\tdBase\tvar\tEE\n";
			//-- create robots, starting randomly at base
			newRobotPositionExtraDistance = 0;
			for (i=0; i<parameters.numOfRobots; i++) {
				CRVector3d position = generateNewRobotPosition();
				Robot robot = new Robot(i, position, CRMaths.getRandomInteger(-180,180), this);
				addRobot(robot);
			}
			
			//-- create deposits
			double grossAmountPerDepo = (double)MAX_DEPOSITS / parameters.numOfDeposits;
			double baseNetAmountPerDepo = (double) MAX_TOTAL_NET_AMOUNT / (grossAmountPerDepo*parameters.numOfDeposits);
			double totalExtraResource = 0;
			ArrayList<Deposit> patchMiddles = new ArrayList<Deposit>();
			
			if (parameters.scenario == World.RANDOM) {
				for (i=0; i< parameters.numOfDeposits; i++) {
					CRVector3d position = generateRandomPositionOutsideOfBase(50);
					Deposit deposit = new Deposit(grossAmountPerDepo, baseNetAmountPerDepo, 100+i, position, new CRVector3d (5,5,0), this);
					addDeposit(deposit);
				}
			} else if (parameters.scenario == World.RANDOM_DISTANT) {
				for (i=0; i< parameters.numOfDeposits; i++) {
					CRVector3d position = generateRandomPositionOutsideOfBase(1500);
					Deposit deposit = new Deposit(grossAmountPerDepo, baseNetAmountPerDepo, 100+i, position, new CRVector3d (5,5,0), this);
					addDeposit(deposit);
				}
			} else if (parameters.scenario == World.PATCHY) {
				int deposPerPatch = 10;
				int deposInCurrentPatch = 0;
				for (i=0; i< parameters.numOfDeposits; i++) {
					CRVector3d position;
					if (deposInCurrentPatch == 0) {
						position = generateRandomPositionOutsideOfBase(500);
						//System.out.println("Patch at " + position.toString());
					} else {
						//-- generate position around 0th deposit in this patch
						position = generateRandomPositionAroundPosition(deposits.get(deposits.size() - deposInCurrentPatch).getPosition(),50);
					}
					Deposit deposit = new Deposit(grossAmountPerDepo, baseNetAmountPerDepo, 100+i, position, new CRVector3d (5,5,0), this);
					addDeposit(deposit);
					if (deposInCurrentPatch == 0) {
						patchMiddles.add(deposit);
					}
					deposInCurrentPatch++;
					//-- move to new patch
					if (deposInCurrentPatch == deposPerPatch) {
						deposInCurrentPatch = 0;
					}
				}
			}
			
			//-- quality locations
			for (i=0; i<parameters.numOfDepositQualityGroups; i++) {
				//-- choose origin, first try from list of patches or otherwise choose at random
				Deposit originDepo;
				if (i < patchMiddles.size()) {
					originDepo = patchMiddles.get(i);
				} else {
					int originDepoId = CRMaths.getRandomInteger(0, parameters.numOfDeposits-1);
					originDepo = deposits.get(originDepoId);
				}
				//System.out.println("ORIGIN: " + originDepo.getId() + "   at " + originDepo.getPosition().toString());
				
				for (Deposit deposit : deposits) {
					double d = Math.abs(deposit.getRelativeVectorTo(originDepo.getPosition()).y); //distance from origin
					double concentration = Math.exp( -(d*d) / (2*parameters.depositQualityVariance*Math.pow(size.x/2,2))) - baseNetAmountPerDepo; // will be adding to baseNetAmountPerDeop so subtract it
					if (concentration > 0) {
						//-- only add amount that will result in total concentration on the depo to be <= 1
						if (concentration + deposit.getResourceConcentration() > 1) {
							concentration -= (concentration + deposit.getResourceConcentration()) - 1;
						}
						deposit.setResourceConcentration(deposit.getResourceConcentration() + concentration);
						totalExtraResource += deposit.getQuantity() * concentration;
						double dToBase = CRMaths.getDistanceOfPoints(deposit.getPosition().x, deposit.getPosition().y, base.getPosition().x, base.getPosition().y);
						double depoEE = Robot.calculateEnergyEfficiency(dToBase, deposit.getQuantity(), deposit.getResourceConcentration());
						info += deposit.getPosition().x + "\t" + deposit.getPosition().y + "\t" + originDepo.getId() + "\t" + deposit.getQuantity() + "\t" + deposit.getResourceConcentration() + "\t" + d + "\t" + dToBase + "\t" + parameters.depositQualityVariance  + "\t"+ depoEE + "\n";
						//System.out.println("setting c at " + deposit.getPosition().toString() + " d = "  + d + "   to " + deposit.getResourceConcentration());
					}
				}
			}
			//System.out.println("TOTAL extra " + totalExtraResource);
			
			//-- count all deposits that haven't had their value lifted
			ArrayList<Deposit> unchangedDepos = new ArrayList<Deposit>();
			for (Deposit deposit : deposits) {
				if (deposit.getResourceConcentration() == baseNetAmountPerDepo) {
					unchangedDepos.add(deposit);
				}
			}
			//-- now, make their concentration slightly lower to compensate for the extra added resource
			int numOfUnchangedDepos = unchangedDepos.size();
			for (Deposit deposit : unchangedDepos) {
				deposit.setResourceConcentration(deposit.getResourceConcentration() - totalExtraResource / (numOfUnchangedDepos*deposit.getQuantity()) );
				double dToBase = CRMaths.getDistanceOfPoints(deposit.getPosition().x, deposit.getPosition().y, base.getPosition().x, base.getPosition().y);
				double depoEE = Robot.calculateEnergyEfficiency(dToBase, deposit.getQuantity(), deposit.getResourceConcentration());
				info += deposit.getPosition().x + "\t" + deposit.getPosition().y + "\t" + -1 + "\t" + deposit.getQuantity() + "\t" + deposit.getResourceConcentration() + "\t" + -1 + "\t" + dToBase + "\t" + parameters.depositQualityVariance + "\t"+ depoEE + "\n";
				//System.out.println(deposit.getResourceConcentration());
			}
			
			//-- check that total is the same as it should be and remember the actual amount of resource
			double totalResource = 0;
			for (Deposit deposit : deposits) {
				totalResource += deposit.getQuantity() * deposit.getResourceConcentration();
				//System.out.println(" > " + totalResource + "   " + deposit.getQuantity() +  "   " + deposit.getResourceConcentration());
			}
			//System.out.println(" COUNTED total resource " + totalResource + "    should be " + MAX_TOTAL_NET_AMOUNT);
			parameters.totalResourceAvailable = totalResource;
			
			//-- print out the resource info
			CRFileIOHelper.stringToFile(info,controller.getReportController().getCurrentFilePath() + "_qualityDeposits");
			
			//-- find out what the base quality is
			baseDepoQuality = 1;
			for (Deposit deposit : deposits) {
				if (deposit.getResourceConcentration() < baseDepoQuality) {
					baseDepoQuality = deposit.getResourceConcentration();
				}
			}
			//System.out.println("Base depo quality: " + baseDepoQuality);
			
			
		} else if (parameters.scenario == TEST) {
			for (i=1; i<=0; i++) {
				Robot robot = new Robot(i, new CRVector3d(CRMaths.getRandomInteger((int)(base.getPosition().x - base.getSize().x/2), (int)(base.getPosition().x + base.getSize().x/2)),
						CRMaths.getRandomInteger((int)(base.getPosition().y - base.getSize().y/2), (int)(base.getPosition().y + base.getSize().y/2)),0),
						CRMaths.getRandomInteger(-180,180), this);
				this.addDynamicModel(robot);
			}
			Robot robot = new Robot(0, new CRVector3d(2000,2000,0), 0, this);
			this.addDynamicModel(robot);
			
			//-- create obstacles
			CRBaseSituatedModel obstacle0 = new CRBaseSituatedModel(10,new CRVector3d(1950,1880,0), new CRVector3d (10,10,0), 0, CRBaseSituatedModel.SHAPE.CIRCLE, CRRenderer.CR_BLACK_DOT);
			this.addSituatedModel(obstacle0);
			CRBaseSituatedModel obstacle0b = new CRBaseSituatedModel(10,new CRVector3d(2030,1930,0), new CRVector3d (10,10,0), 0, CRBaseSituatedModel.SHAPE.CIRCLE, CRRenderer.CR_BLACK_DOT);
			this.addSituatedModel(obstacle0b);
			int min = 1700;
			int max = 2300;
			for (i=min; i< max; i+= 20) {
				CRBaseSituatedModel obstacle = new CRBaseSituatedModel(100 + i,new CRVector3d(i,min,0), new CRVector3d (10,10,0), 0, CRBaseSituatedModel.SHAPE.CIRCLE, CRRenderer.CR_BLACK_DOT);
				this.addSituatedModel(obstacle);
				CRBaseSituatedModel obstacle2 = new CRBaseSituatedModel(300 + i,new CRVector3d(i,max,0), new CRVector3d (10,10,0), 0, CRBaseSituatedModel.SHAPE.CIRCLE, CRRenderer.CR_BLACK_DOT);
				this.addSituatedModel(obstacle2);
				CRBaseSituatedModel obstacle3 = new CRBaseSituatedModel(500 + i,new CRVector3d(min,i,0), new CRVector3d (10,10,0), 0, CRBaseSituatedModel.SHAPE.CIRCLE, CRRenderer.CR_BLACK_DOT);
				this.addSituatedModel(obstacle3);
				CRBaseSituatedModel obstacle4 = new CRBaseSituatedModel(700 + i,new CRVector3d(max,i,0), new CRVector3d (10,10,0), 0, CRBaseSituatedModel.SHAPE.CIRCLE, CRRenderer.CR_BLACK_DOT);
				this.addSituatedModel(obstacle4);
			}
			
			//-- create deposits
			Deposit target0 = new Deposit(1,0.5, 1, new CRVector3d(2000,1900,0), new CRVector3d (5,5,0), this);
			this.addSituatedModel(target0);
			SensorArea area0 = new SensorArea(0,target0.getPosition(),new CRVector3d(120,120,0), CRRenderer.CR_BLANK_CIRCLE);
			this.addSituatedModel(area0);
			
			for (i=1700; i< 2300; i+= 20) {
				Deposit target = new Deposit(1, 0.5, 100 + i, new CRVector3d(i,1800,0), new CRVector3d (5,5,0), this);
				this.addSituatedModel(target);
				Deposit target2 = new Deposit(1, 0.5, 300 + i,new CRVector3d(i,2200,0), new CRVector3d (5,5,0), this);
				this.addSituatedModel(target2);
				Deposit target3 = new Deposit(1, 0.5, 500 + i,new CRVector3d(1800,i,0), new CRVector3d (5,5,0), this);
				this.addSituatedModel(target3);
				Deposit target4 = new Deposit(1, 0.5, 700 + i,new CRVector3d(2200,i,0), new CRVector3d (5,5,0), this);
				this.addSituatedModel(target4);
			}
			
			
			
		}
		
		//-- set initial view to center of base
		if (controller.getRenderer() != null) {
			controller.getRenderer().centerAtViewPosition(new CRVector3d(base.getPosition().x,base.getPosition().y,0));
		}
	}

	public CRVector3d generateNewRobotPosition() {
		CRVector3d pos = new CRVector3d(CRMaths.getRandomInteger((int)(base.getPosition().x - base.getSize().x/2 - newRobotPositionExtraDistance), (int)(base.getPosition().x + base.getSize().x/2 + newRobotPositionExtraDistance)),
				CRMaths.getRandomInteger((int)(base.getPosition().y - base.getSize().y/2 - newRobotPositionExtraDistance), (int)(base.getPosition().y + base.getSize().y/2 + newRobotPositionExtraDistance)),0);
		for (Robot robot : robots) {
			if (CRMaths.getDistanceOfPoints(pos.x, pos.y, robot.getPosition().x, robot.getPosition().y) <= robot.getSize().x + 4) {
				newRobotPositionExtraDistance += 1;
				return generateNewRobotPosition();
			}
		}
		return pos;
	}
	
	public void onTrialEnd(int trialNumber_, int runNumber_) {
		//-- go through robots and write their depo events to a file
		String output = "";
		int counter = 0;
		int numOfEvents = 0;
		for (Robot robot : robots) {
			counter = 0;
			numOfEvents = robot.getFoundDepoEvents().size();
			for (String event : robot.getFoundDepoEvents()) {
				output += event;
				if (counter < numOfEvents - 1) {
					output += ",";
				}
				counter++;
			}
			output += "\n";
		}
		CRFileIOHelper.stringToFile(output, controller.getReportController().getCurrentFilePath() + "_foundDepoEvents");
		super.onTrialEnd(trialNumber_, runNumber_);
		
		clearChildren();
	}

	public CRVector3d generateRandomPositionOutsideOfBase(int minDistanceFromBase_) {
		CRVector3d randPos = new CRVector3d(CRMaths.getRandomInteger(0, (int) this.size.x), CRMaths.getRandomInteger(0,	(int) this.size.y), 0);
		//-- test if it is at least minDistanceFromBase_
		if (base.isAt(randPos, minDistanceFromBase_)) {
			return generateRandomPositionOutsideOfBase(minDistanceFromBase_);
		}
		return randPos;
	}
	
	public CRVector3d generateRandomPositionAroundPosition(CRVector3d position_, int maxDistance_) {
		//-- generate random position from range
		CRVector3d randPos = new CRVector3d(CRMaths.getRandomInteger(0,maxDistance_,true), CRMaths.getRandomInteger(0,maxDistance_,true), 0);
		//-- translate to be around position
		randPos.addVector(position_);
		return randPos;
	}

	public void clearChildren() {
		deposits.clear();
		robots.clear();
		super.clearChildren();
	}

	public void addDeposit(Deposit deposit) {
		deposits.add(deposit);
		this.addSituatedModel(deposit);
	}
	
	public void addRobot(Robot robot) {
		robots.add(robot);
		this.addDynamicModel(robot);
	}

	// -- getters / setters
	public Base getBase() {
		return base;
	}
	public double getBaseDepoQuality() { return baseDepoQuality; }
}
